clc;
clear all;
close all;


%--------------------------------------------------------------------------
%--------------------------------------------------------------------------
%-------------------- Scattered-power Mapping (SPM) -----------------------
%--------------------------------------------------------------------------
%--------------------------------------------------------------------------
% Written by Romina Kazemivala, 1st version: March 25th 2021, 
%                          McMaster University
%--------------------------------------------------------------------------

% This code produces 2D images of the real, imaginary, and magnitude of
% the permittivity for a breast phantom with tumour simulants, of which
% properties is defined on page 22 of the slides

%--------------------------------------------------------------------------
%                               SETTINGS 
%--------------------------------------------------------------------------
% When approx is set to 1 the approximation used is Born's approximation. 
%However, this method can also operate using Rytov's
% approximation when setting approx to 0. Implementation of these
% approximations are described in the "REAL DOMAIN SCATTERED RESPONSE"
%section starting from line 274.

approx=1; % Born is 1, Rytov is 0. 
%--------------------------------------------------------------------------
%Enable Apodization filtering
% NOT REQUIRED FOR SIMULATED EXAMPLES
%Since this example uses experimental data the use of apodization is
%necessary to avoid cross-like artifacts due to the DFT applied in the 
%process of image reconstruction for SPM.
%by setting apod to 1, this filter will be enabled.
%Here a Tukey (tapered cosine) filter is used as an apodization filter and
%can be tuned from 0 to 1,using tukeytol which is refers to "tapers" of 
%the filter. Typically in our experiments we set it to a value between
%0.5-0.8. The apodization section starts from line 250. To view the 
%apodization filter used set show_apod to 1.

apod=1;
%display apodization filter
show_apod=0;
%--------------------------------------------------------------------------
%ENABLE LOW-PASS FILTERING
% Post-inversion low-pass filtering: This filter will alleviate errors
% related to oversampling.Inversion at every k-space position can lead to 
%high-frequency noise in regions it should not exist. In experimental data,
%this filter is a necessity.
%This filter can be enabled by setting lpf to 1. For more inforation on the
%filter setup visit section Low-pass Filter starting from line 499. To view
%the LPF set show_lpf to 1.
% enable low-pass filter setting lpf=1
lpf=1;
%display low-pass filte
show_lpf=0;
%--------------------------------------------------------------------------
%ENABLE FREQUENCY NORMALIZATION
%Frequency normalization is used to get higher resolution images. While 
%electromagnetic waves with lower frequencies penetrate deeper into
%lossy materials and attenuate less, it is only in higher frequencies that
%higher spacial resolution images are obtained. To ensure the
%higher-frequency data is utilized sufficiently, set freq_scaling to 1.
%frequency scaling on the data stage is implemented in line 418.
freq_scaling=1;
%--------------------------------------------------------------------------
%ENABLE PHASE UNWRAP
%When dealing with Rytov's approximated dataset it is better to perform 
%phase unwrapping to avoid discontinuities in the k-space. Frequency
%unwapping can be performed either in the VNAs at the measurement stage 
%or using MATLAB codes on the scattered data. It is unwrapping the phase in
%the real space which is problematic. To enable phase unwrapping set 
%psf_unwrap to 1 when Rytov's approximation is enabled. For more
%information on the process see section Phase Unwrapping section beginning
%from line 318.
psf_unwrap=1;
%--------------------------------------------------------------------------
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%   END OF SETTINGS  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%--------------------------------------------------------------------------

%Scattering parameters are loaded into the code in the INPUT section
%starting from line 122. 
%Note that in this specific experiment the RO data used in approximations 
%for obtaining scattered OUT dataset is different from that used for the 
%SP(CO), therefor, there are two types of RO S-parameters loaded in the 
%INPUT section

%Other variables such as the permittivity of the scattering probe, size of 
%the SP, and the permittivity of the background are defined in the 
%"Define Variables" section starting from line 184. This part is necessary when
%we want to get the estimate of the permittivity after the 2D inverse
%fourier transform is performed on the contrast_func_fft in the 
%"Inverse Problem" section


%--------------------------------------------------------------------------
%                           POWER MAPS                   
%--------------------------------------------------------------------------
%In the SPM method the OUT power map is a projection of the OUT responses 
%in the space of the CO responses. Thus its value at zn is a measure of
%the similarity of the OUT responses and the responses of the CO in which 
%the small scatterer is at zn. The power maps on their own can only
%provide qualitative images. 
%CO power maps are denoted by M_CO and are computed starting from line 438.
%OUT power maps are denoted by M_OUT and are computed starting from line 468.

%--------------------------------------------------------------------------
%                           INVERSION                   
%--------------------------------------------------------------------------
%In order to acquire quantitative results an inverse problem is solved 
%using M_CO and M_OUT. The contrast_func_fft which is the unknown contrast
% function is obtained after the inversion procedure is performed in the 
%Fourier domain.
%eps_out is the quantitative estimate of the complex permittivity in the
%real space of which its imaginary and real part are plotted.
% For the implementation visit the "Inverse Problem" section starting from
% line 563.
%--------------------------------------------------------------------------
%%                          INPUT 
%--------------------------------------------------------------------------
load S11_CO.mat
load S11_RO.mat
load S11_OUT.mat
load S12_CO.mat
load S12_RO.mat
load S12_OUT.mat
load S21_CO.mat
load S21_RO.mat
load S21_OUT.mat
load S22_CO.mat
load S22_RO.mat
load S22_OUT.mat

S11_CO = permute(S11_CO,[1,2,4,3]);
S11_RO = permute(S11_RO,[1,2,4,3]);
S12_CO = permute(S12_CO,[1,2,4,3]);
S12_RO = permute(S12_RO,[1,2,4,3]);
S21_CO = permute(S21_CO,[1,2,4,3]);
S21_RO = permute(S21_RO,[1,2,4,3]);
S22_CO = permute(S22_CO,[1,2,4,3]);
S22_RO = permute(S22_RO,[1,2,4,3]);


%This uses all datasets
% s_CO = {S11_CO;S22_CO;S21_CO;S12_CO};
% s_RO = {S11_RO;S22_RO;S21_RO;S12_RO};
% s_OUT = {S11_OUT;S22_OUT;S21_OUT;S12_OUT};

%this uses only reflection datasets
s_CO = {S11_CO;S22_CO};
s_RO = {S11_RO;S22_RO};
s_OUT = {S11_OUT;S22_OUT};

ftable = 3e9:1e9:8e9;

zz= 30e-3:10e-3:50e-3;

tic

%==========================================================================
%%                    Define Variables
%==========================================================================

c0 = 2.998e8;%Speed of light

[nx, ny, nf,nz]=size(s_CO{1}) %nx, ny and nz are number of samples along x,y and z;
%nf is number of frequency samples
ns = size(s_CO);%ns is the number of responses

frng = 1:nf; %Change this if you want to see less data

% Relative permittivity of background medium
eps_RO = 1;

% Relative permittivity of scattering probe in CO
eps_CO = 1.1;

% Scattering probe size 
vol_sp = (10e-3)^3;%volume of the scattering probe in mm

%Scan Step
xstep = 10e-3;%mm
xrng = xstep*(nx-1);


ROcontr=eps_RO; %relative permittivity of background (RO)

COcontr=eps_CO-eps_RO;% the difference between the relative permittivity of the
%scattering probe and that of the RO)

% if it is a 2D image, zstep is the distance between antennas.
if length(zz)>1
    zstep = zz(2)-zz(1);
else
    zstep = zz(1);
end
%calculating scaling factor for the cubic scattering probe.
scaling_factor=vol_sp/((xstep^2)*zstep); 

%Check to see if the RO dataset is independent for both CO and OUT
%--------------------------------------------------------------------------
if exist('s_RO_forCO','var') == 0
   disp('The RO dataset is the same for both the CO and OUT');
   s_RO_forCO = s_RO;
else
   disp('The RO dataset is independent for both the CO and OUT');
end

 %-------------------------------------------------------------------------
 % Fill out s_RO if it contains only 1 value (i.e. size(s_RO{1} returns (1,1,nz,nf))
 %-------------------------------------------------------------------------
 % in our simulation data, the RO is typically one measurement
% and needs to be built into a map, which is not the case for
            % experimental measurements
if size(s_RO{1}(:,1,1,1)) == 1
    for s = 1:ns
        for f = 1:nf
        s_RO_new{s}(:,:,f) =  (squeeze(s_RO{s}(1,1,f,2))*ones(nx,ny));
        end
    end
end
s_RO = s_RO_new;
%==========================================================================
%%                       Apodization Filter
%==========================================================================
       %This part needs to be applied to the data before Fourier Transform 
       %to diminish cross-like artifacts. This filter is applied in the
       %real space and on the data and is regarded as a pre-processing step
     
if apod==1
    %%%%%%%%%%%%%%%%%%%%%  TUKEY FILTER  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    tukeytol = .75;
    tukey = tukeywin(nx,tukeytol)*tukeywin(nx,tukeytol)';
    apod_filter = tukey;
else
    %if we dont want an apodization filter, just multiply everything by 1!
    apod_filter = ones(nx);
end

%show the apodization filter
if show_apod
    figure;
    imagesc(apod_filter);
    title('Apodization Filter (Tukey Filter 0.75)');
end
  
%==========================================================================
%%               REAL DOMAIN SCATTERED RESPONSE
%==========================================================================
% The scattered response is derived from the CO and OUT responses using
% either Born's or Rytov's approximations
for s = 1:ns
    %----------------------------------------------------------------------
    %                     BORN Approximation
    %----------------------------------------------------------------------
    % Scattered response is S_SC = S_(OUT,CO) - S_RO
    
    if approx
         %loop through each frequency
        for f = 1:nf
            %OUT scattered data
            s_OUT_scat{s}(:,:,f) = s_OUT{s}(:,:,f) - s_RO{s}(:,:,f,1);
            
            %loop through each z-plane
            for z = 1:nz
             %SP(CO) scattered data
                s_CO_scat{s}(:,:,z,f) = s_CO{s}(:,:,f,z) - s_RO_forCO{s}(:,:,f);
                
            end
        end
                            disp('Born approximation');

    %----------------------------------------------------------------------
    %                    RYTOV Approximation
    %----------------------------------------------------------------------

    %Scattered Field OUT or CO = s_RO * ln (s_(OUT or CO)/s_RO)
    else
                            disp('Rytov approximation');
        %OUT scattered data
        s_OUT_scat{s} = s_RO{s}.*log(s_OUT{s}./s_RO{s});
        for z = 1:nz
             %SP(CO) scattered data
            s_CO_scat{s}(:,:,z,:) = s_RO{s}.*log(squeeze(s_CO{s}(:,:,:,z))./s_RO_forCO{s});
        end

    end


end
%==========================================================================
%%              PHASE UNWRAPPING FOR RYTOV'S APPROXIMATION
%==========================================================================    
rng_goodphase = 1:nf;

if approx==0 && psf_unwrap==1
    
    for n = 1:ns
        %Phase unwrapping is a complex procedure for large inhomogeneous
        %objects.
        %Thus, only the real part of s_OUT is used when performing Rytov's
        %approximation to avoid the phase-wrapping issue that appears in
        %the imaginary part (See equation (8) on page 8 of the report 
        
            s_OUT{n} = abs(s_RO{n}).*(log(abs(s_OUT{n})./abs(s_RO{n})));


            %loop through each z plane
            for iz = 1:nz 

            %Interestingly, in some situations for simulated data -infinite
            %data can be generated in the CO measurement. For now equate this
            %to the minimum value.
                
                check_zeros = squeeze(s_CO{n}(:,:,:,iz));                
                min_nonzero = min(abs(check_zeros(check_zeros>0)));                

                 %Phase unwrapping can be performed on the PSF file since 
                 % the scattering probe is small enough so that the field
                 % could be assumed uniform inside it and therefore does
                 % not require complex phase unwrapping procedures.
                 
                 %Here, phase unwrapping is firstly performed along x and
                 %then a second phase unwrapping is applied along y using
                 %MATLAB's unwrap function on the imaginary part of the
                 %s_CO which according to equation (8) of the report is the
                 %difference between CO and RO phases.
                CO_imag_nounwrap = unwrap(unwrap(angle(s_CO{n}(:,:,:)),[],1),[],2) - unwrap(unwrap(angle((s_RO_forCO{n}(:,:,:))),[],1),[],2);
                
                %unwrapped imaginary part of the scattered CO S-parameters
                CO_imagchk{n}(:,:,:,iz) = CO_imag_nounwrap;
                
                
                              
                if isinf(s_CO{n}(:,:,:,iz))
                    s_CO{n}(isinf(s_CO{n}(:,:,:,iz))) = log(min_nonzero./abs(s_RO_forCO{n}(:,:,:)));
                    
                end             
            
            end          
    end 
        
%Range Cleanup
% When we perform Rytov, sometimes we see phase wrapping that cannot be
% corrected. In these cases, we can simply throw out these frequencies. We
% can find which PSFs are affected by finding the gradient of the phase of
% the signal (in this case, the gradient of the signal will show a
% derivative greater than pi/2).
ff_set = [1:51;1:51;1:51;1:51;1:51];

for  n = 1:ns
    for iz =1:nz
        range_update = [];
        for ff = 1:nf
            %finding the gradient of the phase
            gradientmap{n}(:,:,ff,iz) = abs(gradient(gradient(CO_imagchk{n}(:,:,ff,iz)'))');
            badphase_check = max(max(gradientmap{n}(:,:,ff,iz)));
            
           %in case the gradient of the signal is larger than pi/2: contains phase wrapping
            if badphase_check>1.5  
                b = rng_goodphase(rng_goodphase~=ff);% get rid of the problematic frequencies
                rng_goodphase = b;%the range of frequencies that does not include the phase-wrapped data
            end
        end
        
    end
end

%correct the ranges in the scattered data to avoid the phase wrapping
for n = 1:ns
    s_CO{n} = s_CO{n}(:,:,rng_goodphase,:);
    s_OUT{n} = s_OUT{n}(:,:,rng_goodphase);
    s_RO{n} = s_RO{n}(:,:,rng_goodphase);

end

nf = length(rng_goodphase);%the new number of frequency samples after phase-unwrapping
ftable = ftable(rng_goodphase);%%the frequency table after phase-unwrapping
    

end
%==========================================================================
%%                        Fourier Transform
%==========================================================================
%Note that both frequency scaling and apodization filter pre-processing 
%steps are applied prior to the 2D FFT procedure
for s = 1:ns
    
    for f = 1:nf
        
        if freq_scaling %If the frequency normalization is applied on the extracted data
        %prior to the Fourier transform
        maxi = max(max(max(abs(s_CO{s}(:,:,f,:)))));
        else
            maxi=1;%If no frequency scaling is applied just set it to 1.
        end
        %frequency scaling, apodization filter, and the 2D FFT
        %is performed on the OUT scattered data
         s_OUT_fft{s}(:,:,f) = fft2(s_OUT_scat{s}(:,:,f).*apod_filter/maxi);

        for z = 1:nz
            %frequency scaling, apodization filter, and the 2D FFT
            %is performed on the CO scattered data
            s_CO_fft{s}(:,:,z,f) = fft2(s_CO_scat{s}(:,:,z,f).*apod_filter/maxi);

        end
        end
    end


%==========================================================================
%%                   Scattering Probe (SP) Power Maps
%==========================================================================
 % line (9) in Image Reconstruction section of Appendix C

% The CO Maps need to be Nz by Nz at every kx, ky position. 

%First we sum across all S-parameters, performing the conjugate
%multiplication
M_CO_nonorm = zeros(nx,ny,nz,nz,nf);

%for CO we need to loop through z twice, once for the plane being measured (zp),
%and once for the PSF plane using to try and measure that particular plane
%(zn)
for zp = 1:nz
    for zn = 1:nz
        for f=1:nf
            % sum the conjugates across all S-parameters
           for s = 1:ns
              %See Equation (41) on Page 16 of the Report
               M_CO_nonorm(:,:,zp,zn,f) = M_CO_nonorm(:,:,zp,zn,f) + ...
                   s_CO_fft{s}(:,:,zn,f).*conj(s_CO_fft{s}(:,:,zp,f)); 

           end
        end
    end
end
M_CO = sum(M_CO_nonorm,5)/nf;


%==========================================================================
%%                     Generate OUT Power Maps
%==========================================================================

%The OUT Maps need to be Nz by 1 at every kx, ky position. Structing the
%loops to achieve this
M_OUT_nonorm = zeros(nx,ny,nz,nf);
M_OUT_norm = zeros(nx,ny,nz);

%OUT only has one plane being measured, so we loop through all PSFs only
for zn = 1:nz
    %we need to normalize at every frequency
    for f=1:nf
        % sum the conjugates across all S-parameters
        for s = 1:ns
           
            %See Equation (40) on Page 16 of the Report

            M_OUT_nonorm(:,:,zn,f) = M_OUT_nonorm(:,:,zn,f) + ...
                s_OUT_fft{s}(:,:,f).*conj(s_CO_fft{s}(:,:,zn,f));
            
        end
                M_OUT_norm(:,:,zn) = M_OUT_norm(:,:,zn)+ M_OUT_nonorm(:,:,zn,f);

    end
end

% After all the summation we divide by the number of frequencies
M_OUT = M_OUT_norm/nf;
    
     
%==========================================================================
%%                    LOW-PASS FILTER
%==========================================================================
%A Butterworth filter is generated to remove high frequency noise reponses.
%This filter is applied in the Fourier space

%you can try enabling this and multiplying the contrast function

%%%%%%%%%%%%%%%%%%% Butterworth Filter %%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if lpf == 1
   %The highest resolution is limited by the minimum wavelength and the
   %maximum angle that an antenna can see.

%     calculate the minimum wavelength
    lambda_min = c0/sqrt(abs((eps_RO)))/ftable(nf);
    
%     Calculate the maximum angle the antenna can see. This assumes that the
%     directivity of the antenna does not affect the maximum angle, which is
%     typically not the case in experimental reconstructions

       alphaa = atan(sqrt(2*(xrng/7)^2)/zz(nz));
    
%     calculate maximum kmax, and determine is location in the physical
%     FFT images. Beyond kmax the data is primarily noisey.
    kmax = 4*pi*sin(alphaa)/lambda_min;%the equation is given on page 29 of the report
    
    %The FFT generates a spatial frequency range pi/xstep 
    kxx = pi/xstep;
    dkx = 2*kxx/nx;
    kmax_steps = kmax/dkx;
    dev =kmax_steps;%cut-off frequency of the low-pass filter
    
    
%     set the order of the Butterworth filter
    order=4;
    bfilt = ones(nx,ny);
%     BUTTERWORTH FILTER
    for i = 1:nx
        for j =1:ny
            chg = ((i-((nx-1)/2+1))^2 + (j-((ny-1)/2+1))^2)^.5;
%             Create 2D lowpass Butterworth filter.
            bfilt(i,j)= 1/(1 + (chg/dev)^(2*order));
        end
    end
%     Shift the filter to align with the FFT versions of the CO and OUT
    lpf = fftshift(bfilt);
    
%     Any values below 0.01 are automatically set to 0. This reduces
%     computational costs
    lpf = lpf.*(lpf>0.01);
    
else
   % if we dont want to filter, just multiply everything by 1!
    lpf = ones(nx);
end


 %show the Low-pass filter
if show_lpf
    figure;
    imagesc(ifftshift(lpf));
    title('LPF (Butterworth Filter)');
end


%==========================================================================
%%                      INVERSE PROBLEM
%==========================================================================
%Solving a linear system of equations in Equation (42) on Page 16
contrast_func_fft = zeros(nx,ny,nz);
for kx=1:nx
    for ky=1:ny
        %Quantitative Contrast Reconstruction in Fourier Space Step 11 in
        %Apendix C
        contrast_func_fft(kx,ky,:)=pinv(squeeze(M_CO(kx,ky,:,:)))*squeeze(M_OUT(kx,ky,:)); 
    end
end

%Once we have the contrast function (in fourier domain) we can apply our
%scaling factors and generate the permittivity map (eps_out)
eps_out=zeros(nx,ny,nz); 
for z=1:nz
    %Obtain  Real-space  Contrast  via  2D  Inverse  Fourier  Transform
    %Step 13 in Appendix C
    %Note that LPF is applied to the contrast function in the Fourier space
    eps_out(:,:,z)=scaling_factor.*COcontr.*ifftshift(ifft2(lpf.*contrast_func_fft(:,:,z)))+ROcontr; 
end


%==========================================================================
%%                              PLOTS 
%==========================================================================
%Final images are generated here.

if nz==1; %Plot 2D images
for iz = 1:nz

       dx = xstep;
        
       X = (-((nx-1)*dx)/2:dx:((nx-1)*dx)/2).*1000/5;     
    

    fig = figure;
    
%Plot the magnitude of the permittivity
    subplot(1,3,1);
    imagesc(X,X,abs((rot90(eps_out(:,:,iz)))));
    title('Magnitude');
    colorbar;
    ystr = '{\ity} (cm)';
    xstr = '{\itx} (cm)';
    ylabel(ystr);
    xlabel(xstr);
    set(gca,'ydir','normal');
    
%Plot the real part of the permittivity

    subplot(1,3,2);
    imagesc(X,X,real(rot90((eps_out(:,:,iz)))));
    title('Real');
    set(gca,'ydir','normal');
    colorbar;
    ystr = '{\ity} (cm)';
    xstr = '{\itx} (cm)';
    ylabel(ystr);
    xlabel(xstr);
    
%Plot the imaginary part of the permittivity

    subplot(1,3,3);
    imagesc(X,X,imag((rot90(eps_out(:,:,iz)))));
    title('Imaginary');
    colorbar;
    set(gca,'ydir','normal');
    ystr = '{\ity} (cm)';
    xstr = '{\itx} (cm)';
    ylabel(ystr);
    xlabel(xstr);
    
    set(findall(fig,'-property','FontSize'),'FontSize',15)
    set(fig, 'Position', [10, 10, 1200, 300]);
end
else
     % plotting 3D images
    Z=ones(nx,ny);
        [X,Y]=meshgrid(1:nx,1:ny);
        figure
        for k=1:nz
            surface(X,Y,Z.*k,real(eps_out(:,:,k)),'FaceColor','texturemap','CDataMapping','scaled')
        end
        title('Real')
xlabel('x (cm)');
ylabel('y (cm)');
axis tight
view(30,40)
grid on
xh = get(gca,'XLabel'); % Handle of the x label
set(xh, 'Units', 'Normalized')
pos = get(xh, 'Position');
set(xh,'Rotation',25)
yh = get(gca,'YLabel'); % Handle of the y label
set(yh, 'Units', 'Normalized')
pos = get(yh, 'Position');
set(yh,'Rotation',-40)
        colormap parula
        view(3), box on, axis tight square
        shading flat
        set(gca,'YDir','reverse','ZLim',[1 nz])
        colorbar
        
        figure
        for k=1:nz
            surface(X,Y,Z.*k,imag(eps_out(:,:,k)),'FaceColor','texturemap','CDataMapping','scaled')
        end
        title('Imaginary')
        xlabel('x (cm)');
ylabel('y (cm)');
axis tight
view(30,40)
grid on
xh = get(gca,'XLabel'); % Handle of the x label
set(xh, 'Units', 'Normalized')
pos = get(xh, 'Position');
set(xh,'Rotation',25)
yh = get(gca,'YLabel'); % Handle of the y label
set(yh, 'Units', 'Normalized')
pos = get(yh, 'Position');
set(yh,'Rotation',-40)
        colormap parula
        view(3), box on, axis tight square
        shading flat
        set(gca,'YDir','reverse','ZLim',[1 nz])
        colorbar
        
    end

    
 
toc
